#include "Heuristic.h"
#include "Pathfinder.h"
#include <cstdlib>
#include <algorithm>
#include <queue>

#ifdef VISUAL
#include <functional>
#endif

class DifferentialHeuristic: public Heuristic {
public:
	void preprocess(std::vector<int> &map, int width, int height, std::vector<Position> &roots){
		mapWidth = width;
		baseHeuristic->preprocess(map, width, height, roots);
		for(Position p : roots)
			pivots.push_back(p.toIndex(width));
		int nPivots = pivots.size();
		distances.clear();
		distances.resize(map.size() * pivots.size(), std::numeric_limits<double>::infinity());
		std::priority_queue<std::pair<double, int>, std::vector<std::pair<double, int> >,
				std::greater<std::pair<double, int> > > queue;

		for(int i = 0; i < nPivots; i++){
			int pivotIndex = pivots[i];
			distances[pivotIndex * nPivots + i] = 0;
			queue.push(std::make_pair(0, pivotIndex));

			while(!queue.empty()){
				std::pair<double, int> node = queue.top();
				queue.pop();
				if(node.first > distances[node.second * nPivots + i] + EPS)
					continue;

				auto neighbours = Pathfinder::getNeighbours(Position::fromIndex(node.second, width), map, width,
						height);
				for(std::pair<Position, double> neighbour : neighbours){
					double dist = node.first + neighbour.second;
					int nIndex = neighbour.first.toIndex(width);
					if(dist < distances[nIndex * nPivots + i]){
						distances[nIndex * nPivots + i] = dist;
						queue.push(std::make_pair(dist, nIndex));
					}
				}
			}
		}
	}

	double getValue(int nIndex, int gIndex){
		double maxDist = -1;
		double currDist;
		for(size_t i = 0; i < pivots.size(); i++){
			currDist = abs(
					distances[nIndex * pivots.size() + i] - distances[gIndex * pivots.size() + i]);
			if(currDist > maxDist)
				maxDist = currDist;
		}
		return std::max(baseHeuristic->getValue(nIndex, gIndex), maxDist);
	}

	void clear(){
		distances.clear();
		pivots.clear();
	}

	std::string getName(){
		return "Differential Heuristic";
	}

	size_t getSize(){
		return distances.size() * sizeof(double) + (pivots.size() + 1) * sizeof(int) + baseHeuristic->getSize();
	}

	DifferentialHeuristic(std::unique_ptr<Heuristic> base) :
			mapWidth(-1), baseHeuristic(std::move(base)){

	}

private:
	int mapWidth;
	std::unique_ptr<Heuristic> baseHeuristic;
	std::vector<double> distances;
	std::vector<int> pivots;
};
