//============================================================================
// Name        : main.cpp
// Author      : Samuel Bader
// Description : Main methods for benchmarks
//============================================================================

//comment to use GPPC Timer
#define WINDOWS_TIMER

#include <iostream>
#include <vector>
#include <cstdlib>
#include <algorithm>
#include <map>
#include <string>
#include <sstream>
#include <memory>
#include <numeric>
#include <fstream>
#include "ScenarioLoader.h"
#include "Pathfinder.h"
#include "TreeCache.h"
#include "MultipleTrees.h"
#include "SimpleLCA.h"
#include "TreeCacheImproved.h"
#include "AStar.h"
#include "OctileHeuristic.h"
#include "DifferentialHeuristic.h"

#ifndef WINDOWS_TIMER
#include "Timer.h"
#else
#include "WindowsTimer.h"
#endif

void loadMap(const char *filename, std::vector<int> &mapData, int &width, int &height);
double getPathLength(std::vector<Position> &path);
bool validatePath(std::vector<Position> &path, int width, int height,
		const std::vector<int> &mapData);
double calculateRatio(double distance, double expectedDistance);

std::string getCmdOption(int argc, char **argv, const std::string &option,
		const std::string &defaultOption);
bool isCmdOptionSet(int argc, char **argv, const std::string &option);

void saveJsonData(double prepTime, size_t prepSize, std::vector<double> &pathLengths,
		std::vector<double> &runtimes, const std::string &mapName,
		const std::string &pathfinderName);

std::pair<double, size_t> benchmark(std::unique_ptr<Pathfinder>& pathfinder,
		std::vector<int> &mapData, int width, int height, ScenarioLoader &scenario,
		bool printResults,
		std::map<std::string,
				std::pair<std::vector<Position>, std::map<int, std::vector<Position> > > > &pathData,
		std::vector<double> &pathLengths, std::vector<double> &runtimes);

int main(int argc, char **argv){
	bool all = isCmdOptionSet(argc, argv, "--all");

	std::vector<std::pair<std::string, std::string> > mapScenarios;
	std::vector<std::unique_ptr<Pathfinder>> pathfinders;

	if(all){
		setvbuf(stdout, NULL, _IONBF, 0); //Don't Buffer Console output
		mapScenarios.push_back(std::make_pair("AcrosstheCape.map", "AcrosstheCape.map.scen"));
		mapScenarios.push_back(std::make_pair("maze512-1-4.map", "maze512-1-4.map.scen"));
		mapScenarios.push_back(std::make_pair("random512-40-2.map", "random512-40-2.map.scen"));
		mapScenarios.push_back(std::make_pair("rmtst01.map", "rmtst01.map.scen"));

		pathfinders.push_back(std::unique_ptr<Pathfinder>(new TreeCache()));
		pathfinders.push_back(std::unique_ptr<Pathfinder>(new MultipleTrees(16)));
		pathfinders.push_back(std::unique_ptr<Pathfinder>(new SimpleLCA()));
		pathfinders.push_back(std::unique_ptr<Pathfinder>(new TreeCacheImproved(16)));
		pathfinders.push_back(
				std::unique_ptr<Pathfinder>(
						new AStar(1, std::unique_ptr<Heuristic>(new OctileHeuristic()))));
		pathfinders.push_back(
				std::unique_ptr<Pathfinder>(
						new AStar(16,
								std::unique_ptr<Heuristic>(
										new DifferentialHeuristic(
												std::unique_ptr<Heuristic>(
														new OctileHeuristic()))))));
	}else{
		std::string mapFilename = getCmdOption(argc, argv, "-m", "rmtst01.map");
		std::stringstream stream;
		stream << mapFilename << ".scen";
		std::string scenarioFilename = getCmdOption(argc, argv, "-s", stream.str());

		std::unique_ptr<Pathfinder> pathfinder;
		std::string method = getCmdOption(argc, argv, "-p", "tc");
		if(method == "tc"){
			pathfinder = std::unique_ptr<Pathfinder>(new TreeCache());
		}else if(method == "mt" && isCmdOptionSet(argc, argv, "-nr")
				&& isCmdOptionSet(argc, argv, "-nm")){
			int nRootsStart = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			int nRootsEnd = std::stoi(getCmdOption(argc, argv, "-nm", ""));
			for(int i = nRootsStart; i < nRootsEnd - 1; i++)
				pathfinders.push_back(std::unique_ptr<Pathfinder>(new MultipleTrees(i)));
			pathfinder = std::unique_ptr<Pathfinder>(new MultipleTrees(nRootsEnd));
		}else if(method == "mt" && isCmdOptionSet(argc, argv, "-nr")){
			int nRoots = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			pathfinder = std::unique_ptr<Pathfinder>(new MultipleTrees(nRoots));
		}else if(method == "sl"){
			pathfinder = std::unique_ptr<Pathfinder>(new SimpleLCA());
		}else if(method == "tci" && isCmdOptionSet(argc, argv, "-nr")
				&& isCmdOptionSet(argc, argv, "-nm")){
			int nRootsStart = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			int nRootsEnd = std::stoi(getCmdOption(argc, argv, "-nm", ""));
			for(int i = nRootsStart; i < nRootsEnd - 1; i++)
				pathfinders.push_back(std::unique_ptr<Pathfinder>(new TreeCacheImproved(i)));
			pathfinder = std::unique_ptr<Pathfinder>(new TreeCacheImproved(nRootsEnd));
		}else if(method == "tci" && isCmdOptionSet(argc, argv, "-nr")){
			int nRoots = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			pathfinder = std::unique_ptr<Pathfinder>(new TreeCacheImproved(nRoots));
		}else if(method == "all" && isCmdOptionSet(argc, argv, "-nr")){
			int nRoots = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			pathfinders.push_back(std::unique_ptr<Pathfinder>(new MultipleTrees(nRoots)));
			pathfinders.push_back(std::unique_ptr<Pathfinder>(new TreeCacheImproved(nRoots)));
			pathfinder = std::unique_ptr<Pathfinder>(
					new AStar(nRoots,
							std::unique_ptr<Heuristic>(
									new DifferentialHeuristic(
											std::unique_ptr<Heuristic>(new OctileHeuristic())))));
		}else if(method == "all"){
			pathfinders.push_back(std::unique_ptr<Pathfinder>(new TreeCache()));
			pathfinders.push_back(
					std::unique_ptr<Pathfinder>(
							new AStar(1, std::unique_ptr<Heuristic>(new OctileHeuristic()))));
			pathfinder = std::unique_ptr<Pathfinder>(new SimpleLCA());
		}else if(method == "asoc"){
			pathfinder = std::unique_ptr<Pathfinder>(
					new AStar(1, std::unique_ptr<Heuristic>(new OctileHeuristic())));
		}else if(method == "asdh" && isCmdOptionSet(argc, argv, "-nr")){
			int nRoots = std::stoi(getCmdOption(argc, argv, "-nr", ""));
			pathfinder = std::unique_ptr<Pathfinder>(
					new AStar(nRoots,
							std::unique_ptr<Heuristic>(
									new DifferentialHeuristic(
											std::unique_ptr<Heuristic>(new OctileHeuristic())))));
		}else{
			std::cout << "Invalid Arguments" << std::endl << "Valid Arguments: "
					<< "[--all] [-p <pathfinder> [-nr <nRoots> [-nm <maxNRoots>]]] "
					<< "[--savePaths [-rs <startIndex>] [-re <endIndex>]] "
					<< "[-m <map> -s <scenario>] [--json]";
			exit(0);
		}

		mapScenarios.push_back(std::make_pair(mapFilename, scenarioFilename));
		pathfinders.push_back(std::move(pathfinder));
	}

	bool saveJson = isCmdOptionSet(argc, argv, "--json");

	std::vector<int> mapData;
	int width, height;

	for(auto mapScenario : mapScenarios){
		std::stringstream stream;
		stream << "maps\\" << mapScenario.first;
		loadMap(stream.str().c_str(), mapData, width, height);

		stream.str("");
		stream << "maps\\" << mapScenario.second;
		ScenarioLoader scenario(stream.str().c_str());
		int nExperiments = scenario.GetNumExperiments();

		bool savePaths = isCmdOptionSet(argc, argv, "--savePaths");
		int startIndex = 0, endIndex = nExperiments;
		std::map<std::string,
				std::pair<std::vector<Position>, std::map<int, std::vector<Position> > > > pathData;
		stream.str("");
		stream << "maps\\" << mapScenario.first << ".csv";
		std::string pathFilename = stream.str();

		if(savePaths){
			if(isCmdOptionSet(argc, argv, "-rs"))
				startIndex = std::stoi(getCmdOption(argc, argv, "-rs", ""));
			if(isCmdOptionSet(argc, argv, "-re"))
				endIndex = std::stoi(getCmdOption(argc, argv, "-re", ""));
			if(startIndex > endIndex || startIndex >= nExperiments)
				startIndex = 0;
			if(endIndex >= nExperiments)
				endIndex = nExperiments;

			std::string line;
			std::ifstream pathFile(pathFilename);
			if(pathFile.is_open()){
				std::getline(pathFile, line, ',');
				int nMethods = stoi(line);
				std::getline(pathFile, line);
				int nPaths = stoi(line);
				for(int k = 0; k < nMethods; k++){
					std::map<int, std::vector<Position> > pathMap;
					std::string method;
					std::getline(pathFile, method);
					std::vector<Position> roots;
					std::getline(pathFile, line, ','); //roots
					std::getline(pathFile, line, ','); //l
					int nRoots = stoi(line);
					for(int j = 0; j < nRoots; j++){
						std::getline(pathFile, line, ','); //x
						int x = stoi(line);
						std::getline(pathFile, line, ','); //y
						int y = stoi(line);
						roots.push_back(Position(x, y));
					}
					std::getline(pathFile, line); //EOL
					for(int i = 0; i < nPaths; i++){
						std::vector<Position> path;
						std::getline(pathFile, line, ' '); //path
						std::getline(pathFile, line, ','); //index
						int index = stoi(line);
						std::getline(pathFile, line, ','); //l
						int l = stoi(line);
						for(int j = 0; j < l; j++){
							std::getline(pathFile, line, ','); //x
							int x = stoi(line);
							std::getline(pathFile, line, ','); //y
							int y = stoi(line);
							path.push_back(Position(x, y));
						}
						pathMap[index] = path;
						std::getline(pathFile, line); //EOL
					}
					if(pathMap.find(startIndex) != pathMap.end()
							&& pathMap.find(endIndex - 1) != pathMap.end())
						pathData[method] = std::make_pair(roots, pathMap);
				}
				pathFile.close();
			}
		}

		for(std::unique_ptr<Pathfinder>& pathfinder : pathfinders){
			if(all)
				std::cout << "Benchmarking for " << mapScenario.second << " with "
						<< pathfinder->getName() << "...";
			std::vector<double> pathLengths, runtimes;
			auto preprocessData = benchmark(pathfinder, mapData, width, height, scenario, !all,
					pathData, pathLengths, runtimes);
			if(saveJson)
				saveJsonData(preprocessData.first, preprocessData.second, pathLengths, runtimes,
						mapScenario.first, pathfinder->getName());
			if(all)
				std::cout << "done, preprocessing time: " << preprocessData.first
						<< ", total search time: "
						<< std::accumulate(runtimes.begin(), runtimes.end(), 0.0) << std::endl;
		}
		if(savePaths){
			std::ofstream pathFile(pathFilename);
			int saveNExperiments = endIndex - startIndex;
			pathFile << pathData.size() << "," << saveNExperiments << std::endl;
			for(auto dataPair : pathData){
				pathFile << dataPair.first << std::endl;
				std::vector<Position> roots = dataPair.second.first;
				pathFile << "roots,";
				pathFile << roots.size() << ",";
				for(Position pos : roots){
					pathFile << pos.x << "," << pos.y << ",";
				}
				pathFile << std::endl;
				std::vector<Position> path;
				for(int i = startIndex; i < endIndex; i++){
					path = dataPair.second.second[i];
					pathFile << "path " << i << ",";
					pathFile << path.size() << ",";
					for(Position pos : path){
						pathFile << pos.x << "," << pos.y << ",";
					}
					pathFile << std::endl;
				}
			}
			pathFile.close();
		}

	}

	return 0;
}

std::pair<double, size_t> benchmark(std::unique_ptr<Pathfinder>& pathfinder,
		std::vector<int> &mapData, int width, int height, ScenarioLoader &scenario,
		bool printResults,
		std::map<std::string,
				std::pair<std::vector<Position>, std::map<int, std::vector<Position> > > > &pathData,
		std::vector<double> &pathLengths, std::vector<double> &runtimes){

#ifndef WINDOWS_TIMER
	Timer t;
#else
	WindowsTimer t;
#endif
	double preprocessTime;
	int nExperiments = scenario.GetNumExperiments();
	pathLengths.resize(nExperiments);
	runtimes.resize(nExperiments);
	t.StartTimer();
	pathfinder->preprocess(mapData, width, height);
	t.EndTimer();
	preprocessTime = t.GetElapsedTime();
	size_t preprocessSize = pathfinder->getSize();
	if(printResults)
		std::cout << "Preprocessing time[mus]:" << preprocessTime << ", size: " << preprocessSize
				<< ", roots:" << pathfinder->getRoots().size() << std::endl;

	std::map<int, std::vector<Position> > paths;
	pathData.erase(pathfinder->getName());
	for(int i = 0; i < nExperiments; ++i){
		std::vector<Position> path;
		Position start(scenario.GetNthExperiment(i).GetStartX(),
				scenario.GetNthExperiment(i).GetStartY());
		Position goal(scenario.GetNthExperiment(i).GetGoalX(),
				scenario.GetNthExperiment(i).GetGoalY());
		t.StartTimer();
		pathfinder->findPath(start, goal, path, mapData, width, height);
		t.EndTimer();
		pathLengths[i] = getPathLength(path);
		runtimes[i] = t.GetElapsedTime();
		if(printResults){
			printf("id\t%d\ttime[mus]\t%f\tdist\t%f\texp-dist\t%f\tsubopt\t%f\t", i, runtimes[i],
					pathLengths[i], scenario.GetNthExperiment(i).GetDistance(),
					calculateRatio(getPathLength(path),
							scenario.GetNthExperiment(i).GetDistance()));
			if((path.size() == 0 && scenario.GetNthExperiment(i).GetDistance() == 0)
					|| (path.size() != 0 && validatePath(path, width, height, mapData)
							&& start.x == path[0].x && start.y == path[0].y
							&& goal.x == path.back().x && goal.y == path.back().y)){
				printf("valid\n");
			}else{
				printf("invalid\n");
			}
			if(path.size() == 0){
				path.push_back(Position(start.x, start.y));
				path.push_back(Position(goal.x, goal.y));
			}
			paths[i] = path;
		}
	}
	pathData[pathfinder->getName()] = std::make_pair(pathfinder->getRoots(), paths);
	pathfinder->clear();
	return std::make_pair(preprocessTime, preprocessSize);
}

double getPathLength(std::vector<Position> &path){
	double len = 0;
	for(int i = 0; i < (int) path.size() - 1; i++){
		if(path[i].x == path[i + 1].x || path[i].y == path[i + 1].y){
			len++;
		}else{
			len += DIAG_DIST;
		}
	}
	return len;
}

bool validatePath(std::vector<Position> &path, int width, int height,
		const std::vector<int> &mapData){
	for(int i = 0; i < (int) path.size() - 1; i++){
		if(abs(path[i].x - path[i + 1].x) > 1)
			return false;
		if(abs(path[i].y - path[i + 1].y) > 1)
			return false;
		if(!mapData[path[i].y * width + path[i].x])
			return false;
		if(!mapData[path[i + 1].y * width + path[i + 1].x])
			return false;
		if(path[i].x != path[i + 1].x && path[i].y != path[i + 1].y){
			if(!mapData[path[i + 1].y * width + path[i].x])
				return false;
			if(!mapData[path[i].y * width + path[i + 1].x])
				return false;
		}
	}
	return true;
}

/*information from http://www.movingai.com/benchmarks/formats.html:
 All maps begin with the lines:

 type octile
 height x
 width y
 map

 where x and y are the repsective height and width of the map.

 The map data is store as an ASCII grid. The following characters are possible:

 . - passable terrain
 G - passable terrain
 @ - out of bounds
 O - out of bounds
 T - trees (unpassable)
 S - swamp (passable from regular terrain)
 W - water (traversable, but not passable from terrain)

 */
//taken from http://movingai.com/GPPC/code.html
void loadMap(const char *filename, std::vector<int> &mapData, int &width, int &height){
	FILE *f;
	f = fopen(filename, "r");
	if(f){
		fscanf(f, "type octile\nheight %d\nwidth %d\nmap\n", &height, &width);
		mapData.resize(height * width);
		for(int y = 0; y < height; y++){
			for(int x = 0; x < width; x++){
				char c;
				do{
					fscanf(f, "%c", &c);
				}while(isspace(c));
				mapData[y * width + x] = (c == '.' || c == 'G' || c == 'S') ? 1 : 0;
				//printf("%c", c);
			}
			//printf("\n");
		}
		fclose(f);
	}else{
		std::cout << "Map load fail" << std::endl;
	}
}

double calculateRatio(double distance, double expectedDistance){
	double ratio = distance / expectedDistance;
	if(ratio < 1){
		if(expectedDistance - distance < 1) //rounding in .scen files
			return 1;
	}
	return ratio;
}

void saveJsonData(double prepTime, size_t prepSize, std::vector<double> &pathLengths,
		std::vector<double> &runtimes, const std::string &mapName,
		const std::string &pathfinderName){
	std::ofstream jsonFile("results\\properties", std::ios_base::app);
	jsonFile << "{";
	jsonFile << "\"" << pathfinderName << "_" << mapName << "_prep\": {" << std::endl;
	jsonFile << "    \"domain\": \"" << mapName << "\"," << std::endl;
	jsonFile << "    \"problem\": \"prep\"," << std::endl;
	jsonFile << "    \"config\": \"" << pathfinderName << "\"," << std::endl;
	jsonFile << "    \"size\": " << prepSize << "," << std::endl;
	jsonFile << "    \"time\": " << prepTime << std::endl;
	jsonFile << "}";
	for(unsigned int i = 0; i < pathLengths.size(); i++){
		jsonFile << "," << std::endl;
		jsonFile << "\"" << pathfinderName << "_" << mapName << "_" << i << "\": {" << std::endl;
		jsonFile << "    \"domain\": \"" << mapName << "\"," << std::endl;
		jsonFile << "    \"problem\": \"" << i << "\"," << std::endl;
		jsonFile << "    \"config\": \"" << pathfinderName << "\"," << std::endl;
		jsonFile << "    \"length\": " << pathLengths[i] << "," << std::endl;
		jsonFile << "    \"time\": " << runtimes[i] << std::endl;
		jsonFile << "}";
	}
	jsonFile << "}";
	jsonFile.close();
}

std::string getCmdOption(int argc, char **argv, const std::string &option,
		const std::string &defaultOption){
	char **end = argv + argc;
	char **itr = std::find(argv, end, option);
	if(itr != end && ++itr != end){
		return *itr;
	}
	return defaultOption;
}

bool isCmdOptionSet(int argc, char **argv, const std::string &option){
	char **end = argv + argc;
	return std::find(argv, end, option) != end;
}
